import sys, random, math
import numpy as np
from scipy import stats, spatial, sparse, linalg
from collections import defaultdict
from sklearn.preprocessing import normalize

def ycn_edges(bipartite, keys, directed = False):
   ### This part just to build the dense matrix T (that is list of lists)
   projected_keys = sorted([key for key in keys if keys[key] == 0])
   other_keys = sorted([key for key in keys if keys[key] != 0])
   T = []
   for i in range(len(projected_keys)):
      I = []
      for j in range(len(other_keys)):
         if other_keys[j] in bipartite[projected_keys[i]]:
            I.append(1.0)
         else:
            I.append(0.0)
      T.append(I)
   ### End T building
   T = sparse.csr_matrix(T) # Convert T to compressed sparse row matrix (for space efficiency)
   T_norm = normalize(T, norm = 'l1', axis = 1) # Divide each row element with the row sum (Eq. [1] in the paper)
   T_t_norm = normalize(T.T, norm = 'l1', axis = 1) # Divide each row element of the transpose with the transposed row sum (Eq. [2] in the paper) 
   T = T_norm.dot(T_t_norm) # Multiply (Eq. [3] in the paper)
   size = T.shape[0]
   ### Calculation of stationary distribution (pi)
   P = T
   pi = np.ones(size)/size # Of course, a vector of zeroes
   pi1 = np.zeros(size)
   n = linalg.norm(pi - pi1, 1) # This puts in n the maximum absolute value of the difference between pi and pi1
   i = 0
   while n > 1e-9 and i < 1e5: # Keep cycling until the difference (n) is minimal, or you cycled a lot
      pi1 = pi * P
      pi = pi1 * P
      n = linalg.norm(pi - pi1, 1) # As before, it puts in n the maximum absolute value of the difference between pi and pi1
      i += 1
   ### End calculation of pi
   edges = defaultdict(int)
   indices_tocheck = T.nonzero()
   for i in range(len(indices_tocheck[0])): # Cycle over all non.zero elements of T
      x = indices_tocheck[0][i]
      y = indices_tocheck[1][i]
      if directed:
         edges[(projected_keys[x], projected_keys[y])] = T[x, y] * pi[x]
      else:
         if projected_keys[x] < projected_keys[y]: # For undirected networks, I always check that the first element has a label smaller than the second, for directed networks we'll ignore this
            edges[(projected_keys[x], projected_keys[y])] = max(T[x, y] * pi[x], T[y, x] * pi[y]) # x and y have always two similarities: Txy times pix and Tyx times piy. I always take the maximum of the two.
         elif projected_keys[x] > projected_keys[y]:
            edges[(projected_keys[y], projected_keys[x])] = max(T[x, y] * pi[x], T[y, x] * pi[y])
   return edges

def jaccard(bipartite, keys):
   edges = defaultdict(int)
   projected_keys = [key for key in keys if keys[key] == 0]
   for i in range(len(projected_keys) - 1):
      set_i = set(bipartite[projected_keys[i]])
      for j in range(i + 1, len(projected_keys)):
         set_j = set(bipartite[projected_keys[j]])
         if len(set_i | set_j) > 0:
            if projected_keys[i] < projected_keys[j]:
               edges[(projected_keys[i], projected_keys[j])] = float(len(set_i & set_j)) / len(set_i | set_j)
            elif projected_keys[i] > projected_keys[j]:
               edges[(projected_keys[j], projected_keys[i])] = float(len(set_i & set_j)) / len(set_i | set_j)
   return edges

def euclidean(bipartite, keys):
   edges = defaultdict(int)
   projected_keys = [key for key in keys if keys[key] == 0]
   other_keys = [key for key in keys if keys[key] != 0]
   binarized = {}
   for i in range(len(projected_keys)):
      list_i = []
      for y in range(len(other_keys)):
         if other_keys[y] in bipartite[projected_keys[i]]:
            list_i.append(1)
         else:
            list_i.append(0)
      binarized[projected_keys[i]] = list_i
   for i in range(len(projected_keys) - 1):
      for j in range(i + 1, len(projected_keys)):
         ij_distance = spatial.distance.euclidean(binarized[projected_keys[i]], binarized[projected_keys[j]])
         if ij_distance == 0:
            ij_distance = 1.0
         if projected_keys[i] < projected_keys[j]:
            edges[(projected_keys[i], projected_keys[j])] = 1.0 / ij_distance
         elif projected_keys[i] > projected_keys[j]:
            edges[(projected_keys[j], projected_keys[i])] = 1.0 / ij_distance
   return edges

def cosine(bipartite, keys):
   edges = defaultdict(int)
   projected_keys = [key for key in keys if keys[key] == 0]
   other_keys = [key for key in keys if keys[key] != 0]
   binarized = {}
   for i in range(len(projected_keys)):
      list_i = []
      for y in range(len(other_keys)):
         if other_keys[y] in bipartite[projected_keys[i]]:
            list_i.append(1)
         else:
            list_i.append(0)
      binarized[projected_keys[i]] = list_i
   for i in range(len(projected_keys) - 1):
      for j in range(i + 1, len(projected_keys)):
         ij_distance = spatial.distance.cosine(binarized[projected_keys[i]], binarized[projected_keys[j]])
         if math.isnan(ij_distance):
            ij_distance = 1.0
         if projected_keys[i] < projected_keys[j]:
            edges[(projected_keys[i], projected_keys[j])] = 1.0 - ij_distance
         elif projected_keys[i] > projected_keys[j]:
            edges[(projected_keys[j], projected_keys[i])] = 1.0 - ij_distance
   return edges

def pearson(bipartite, keys):
   edges = defaultdict(int)
   projected_keys = [key for key in keys if keys[key] == 0]
   other_keys = [key for key in keys if keys[key] != 0]
   binarized = {}
   for i in range(len(projected_keys)):
      list_i = []
      for y in range(len(other_keys)):
         if other_keys[y] in bipartite[projected_keys[i]]:
            list_i.append(1)
         else:
            list_i.append(0)
      binarized[projected_keys[i]] = list_i
   for i in range(len(projected_keys) - 1):
      for j in range(i + 1, len(projected_keys)):
         ij_distance = spatial.distance.correlation(binarized[projected_keys[i]], binarized[projected_keys[j]])
         if projected_keys[i] < projected_keys[j]:
            edges[(projected_keys[i], projected_keys[j])] = (2.0 - ij_distance) / 2.0
         elif projected_keys[i] > projected_keys[j]:
            edges[(projected_keys[j], projected_keys[i])] = (2.0 - ij_distance) / 2.0
   return edges

def zhou(bipartite, keys, directed = False):
   projected_keys = [key for key in keys if keys[key] == 0]
   other_keys = [key for key in keys if keys[key] != 0]
   T = []
   for i in range(len(projected_keys)):
      I = []
      for j in range(len(other_keys)):
         if other_keys[j] in bipartite[projected_keys[i]]:
            I.append(1.0)
         else:
            I.append(0.0)
      T.append(I)
   T = sparse.csr_matrix(T)
   T_norm = normalize(T, norm = 'l1', axis = 1)
   T_t_norm = normalize(T.T, norm = 'l1', axis = 1)
   T = T_norm.dot(T_t_norm)
   edges = defaultdict(int)
   indices_tocheck = T.nonzero()
   for i in range(len(indices_tocheck[0])):
      x = indices_tocheck[0][i]
      y = indices_tocheck[1][i]
      if directed:
         edges[(projected_keys[x], projected_keys[y])] = T[x, y]
      else:
         if projected_keys[x] < projected_keys[y]:
            edges[(projected_keys[x], projected_keys[y])] = max(T[x, y], T[y, x])
         elif projected_keys[x] > projected_keys[y]:
            edges[(projected_keys[y], projected_keys[x])] = max(T[x, y], T[y, x])
   return edges
